/* 
 */

/*

    Copyright (C) 2014 Ferrero Andrea

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.


 */

/*

    These files are distributed with PhotoFlow - http://aferrero2707.github.io/PhotoFlow/

 */

#include <glibmm/iochannel.h>

#include "photoflow.hh"
#include "options.hh"
#include "../rt/rtengine/safekeyfile.h"

PF::Options::Options()
{
  display_profile_type = PF::PF_DISPLAY_PROF_sRGB;
  save_sidecar_files = 0;
}

void PF::Options::set_display_profile_type(int t)
{
  if( t>= 0 && t < PF::PF_DISPLAY_PROF_MAX)
    display_profile_type = (display_profile_t)t;
}


void PF::Options::load()
{
  setlocale(LC_NUMERIC, "C"); // to set decimal point to "."
  rtengine::SafeKeyFile keyFile;

  Glib::ustring fname =
      Glib::build_filename( Glib::ustring(PF::PhotoFlow::Instance().get_config_dir()), "options" );
  std::cout<<"Loading custom settings..."<<std::endl;
  try {
    if (keyFile.load_from_file (fname)) {

      // --------------------------------------------------------------------------------------------------------

      if (keyFile.has_group ("Folders")) {
        if (keyFile.has_key ("Folders", "last_visited_image_folder")) {
          last_visited_image_folder = keyFile.get_string ("Folders", "last_visited_image_folder");
        }
        if (keyFile.has_key ("Folders", "last_visited_preset_folder")) {
          last_visited_preset_folder = keyFile.get_string ("Folders", "last_visited_preset_folder");
        }
        if (keyFile.has_key ("Folders", "last_visited_icc_folder")) {
          last_visited_icc_folder = keyFile.get_string ("Folders", "last_visited_icc_folder");
        }
      }

      if (keyFile.has_group ("Output")) {
        if (keyFile.has_key ("Output", "save_sidecar_files")) {
          save_sidecar_files = keyFile.get_integer ("Output", "save_sidecar_files");
        }
      }

      if (keyFile.has_group ("Color Management")) {
        if (keyFile.has_key ("Color Management", "DisplayProfileType")) {
          int keyval = keyFile.get_integer ("Color Management", "DisplayProfileType");
          if( keyval>0 && keyval<PF::PF_DISPLAY_PROF_MAX)
            display_profile_type = (display_profile_t)keyval;
          std::cout<<"display_profile_type="<<display_profile_type<<std::endl;
        }
        if (keyFile.has_key ("Color Management", "CustomDisplayProfileName")) {
          custom_display_profile_name = keyFile.get_string ("Color Management", "CustomDisplayProfileName");
          std::cout<<"custom_display_profile_name="<<custom_display_profile_name<<std::endl;
        }
      }
    }
  } catch (Glib::Error &err) {
    printf("Options::readFromFile / Error code %d while reading values from \"%s\":\n%s\n", err.code(), fname.c_str(), err.what().c_str());
  } catch (...) {
    printf("Options::readFromFile / Unknown exception while trying to load \"%s\"!\n", fname.c_str());
  }
  std::cout<<"... custom settings loaded."<<std::endl; //getchar();
  std::cout<<"get_save_sidecar_files(): "<<get_save_sidecar_files()<<std::endl;
}


void PF::Options::save()
{
  setlocale(LC_NUMERIC, "C"); // to set decimal point to "."
  rtengine::SafeKeyFile keyFile;

  Glib::ustring fname =
      Glib::build_filename( Glib::ustring(PF::PhotoFlow::Instance().get_config_dir()), "options" );

  keyFile.set_string ("Folders", "last_visited_image_folder", last_visited_image_folder);
  keyFile.set_string ("Folders", "last_visited_preset_folder", last_visited_preset_folder);
  keyFile.set_string ("Folders", "last_visited_icc_folder", last_visited_icc_folder);

  keyFile.set_integer ("Output", "save_sidecar_files", (int)save_sidecar_files);

  keyFile.set_integer ("Color Management", "DisplayProfileType", (int)display_profile_type);
  keyFile.set_string ("Color Management", "CustomDisplayProfileName", custom_display_profile_name);

  try {
    //keyFile.save_to_file( fname );
    Glib::ustring opt_str = keyFile.to_data();
    //std::string fnames = fname.get_raw();
    std::cout<<"Saving options..."<<std::endl;
    std::cout<<opt_str<<std::endl;
    std::string mode = "w";
    Glib::RefPtr< Glib::IOChannel > ioch = Glib::IOChannel::create_from_file( fname.raw(), mode );
    ioch->write( opt_str );
    ioch->close();
    std::cout<<"... options saved."<<std::endl;
  } catch (Glib::Error &err) {
    printf("Options::readFromFile / Error code %d while saving values to \"%s\":\n%s\n", err.code(), fname.c_str(), err.what().c_str());
  } catch (...) {
    printf("Options::readFromFile / Unknown exception while trying to save \"%s\"!\n", fname.c_str());
  }
}
